﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace LinearSolver
{
    public class LinearSystem
    {
        private readonly string Share = "C:\\Devs\\Output\\SAT";
        private IList<Equation> BinarySystem = new List<Equation>();
        private IList<Equation> SecondaryBinarySystem = new List<Equation>();
        private IDictionary<int, List<int>> map = new Dictionary<int, List<int>>();
        public bool ShowLog = false;
        public int Count
            
            {
    get { return BinarySystem.Count;; }}
    
        public void Clear()
        {
            BinarySystem.Clear();
        }

        public void AddBinarySystem(HashSet<Equation> bs)
        {
            foreach (Equation eq in bs)
                BinarySystem.Add(eq);
        }
        public void Add(Equation eq)
        {
            BinarySystem.Add(eq);
        }
        public void LoadLS()
        {
            string FileName = Path.Combine(Share, "LS.txt");
            BinarySystem.Clear();
            int count = 0;
            using (StreamReader sr = new StreamReader(FileName))
            {
                string line;
                while ((line = sr.ReadLine()) != null)
                {
                    Equation eq = new Equation();
                    string[] s = line.Split('|');
                    eq.Value = int.Parse(s[0]);
                    string[] ss = s[1].Split(' ');
                    foreach (string v in ss)
                    {
                        if (string.IsNullOrEmpty(v)) continue;
                        int a = int.Parse(v);
                        eq.Add(a);
                        if (map.ContainsKey(a))
                            map[a].Add(count);
                        else
                        {
                            map.Add(a, new List<int>() { count });
                        }
                    }
                    count++;
                    BinarySystem.Add(eq);
                }
            }
        }

        private void DumpLS(int n = 0)
        {
            string FileName;
            if (n == 0)
                FileName = Path.Combine(Share, "LS.txt");
            else
                FileName = Path.Combine(Share, "LS1.txt");

            using (StreamWriter sw = new StreamWriter(FileName))
            {
                string line;
                foreach (Equation eq in BinarySystem)
                {
                    line = eq.GetLine();
                    sw.WriteLine(line);
                }
            }
        }
        public void generateBinarySystem(int NbrVariables, int NbrClauses)
        {
            Directory.CreateDirectory(Share);
            Random _random = new Random();

            for (int i = 0; i < NbrClauses; i++)
            {
                int var = _random.Next(0, 2);
                Equation eq = new Equation();
                eq.Value = var;
                int n = _random.Next(3, NbrVariables);

                for (int j = 0; j < n; j++)
                {
                    int a = _random.Next(0, n);
                    eq.Add(a);

                }
                BinarySystem.Add(eq);
            }

            // Dump system
            DumpLS();
        }
        public void UpdateSystem(int N , int NN)
        {
            foreach (Equation eq in BinarySystem)
                eq.UpdateEquation(N , NN);
        }
        //  if a variable appears once ( in only one equation ) we can remove the equation from the system.
        private IDictionary<int, int> SplitSystem()
        {
            IDictionary<int, int> dic = new Dictionary<int, int>();
            for (int i = 0; i < BinarySystem.Count; i++)
            {
                Equation eq = BinarySystem[i];
                foreach (int v in eq.Variables)
                {
                    if (dic.ContainsKey(v))
                    {
                        dic[v] += 1 ;
                    }
                    else
                    {
                        dic.Add(v, 1);
                    }

                }

            }

            SecondaryBinarySystem.Clear();
            // split binary system;
            foreach (Equation eq in BinarySystem.Reverse())
            {
                foreach (int v in eq.Variables)
                {
                    if (dic[v]==1) // only one equation contains the current variable.
                    {
                        SecondaryBinarySystem.Add(eq);
                        BinarySystem.Remove(eq);
                        break;

                    }
                }
            }

            return dic;
        }
        public IDictionary<int, int> SolveP()
        {
            int i = 0;
            for (int k = 0; k < BinarySystem.Count; k++)
            {
                Equation eq = BinarySystem[k];
                int j = -1;
                eq.Prepare();
                int a = eq.Pivot;
                var events = new List<ManualResetEvent>();

                for (int l = 0; l < BinarySystem.Count; l++)
                {
                    j++;
                    Equation ww = BinarySystem[l];
                    if (i >= j) continue;
                    var resetEvent = new ManualResetEvent(false);
                    ThreadPool.QueueUserWorkItem(
                    arg =>
                            {
                                ww.Propagate(eq);
                                resetEvent.Set();
                            });
                    events.Add(resetEvent);

                }

                var wait = true;
                while (wait)
                {
                    if (events.Count > 0)
                    {
                        WaitHandle.WaitAll(events.Take(60).ToArray());
                        events.RemoveRange(0, events.Count > 59 ? 60 : events.Count);
                        wait = events.Any();
                    }
                    else
                        wait = false;
                }

                if (ShowLog)
                {
                    //if (i % 100 == 0)
                    //    Console.WriteLine($"{i} items handled time {DateTime.Now}");

                    //if (i == BinarySystem.Count - 1)
                    //    Console.WriteLine($"{i} items handled time {DateTime.Now}");
                }
                i++;
            }

            // Dump system
            if(ShowLog)
                DumpLS(1);

            // return solution
            IDictionary<int, int> solutions =GetASolution();
            return solutions;
        }


        public IDictionary<int, int> SolveP2()
        {
            int i = 0;
            for (int k = 0; k < BinarySystem.Count; k++)
            {
                Equation eq = BinarySystem[k];
                int j = -1;
                eq.Prepare();
                int a = eq.Pivot;
                var events = new List<ManualResetEvent>();
                IList<Task> tsks = new List<Task>();


                for (int l = 0; l < BinarySystem.Count; l++)
                {
                    j++;
                    Equation ww = BinarySystem[l];
                    if (i >= j) continue;

                    var task = Task.Factory.StartNew(() => ww.Propagate(eq));
                    tsks.Add(task);
                }
                Parallel.ForEach(tsks, tsk => tsk.Wait());

                if (ShowLog)
                {
                    //if (i % 100 == 0)
                    //    Console.WriteLine($"{i} items handled time {DateTime.Now}");

                    //if (i == BinarySystem.Count - 1)
                    //    Console.WriteLine($"{i} items handled time {DateTime.Now}");
                }
                i++;
            }

            // Dump system
            if (ShowLog)
                DumpLS(1);

            // return solution
            IDictionary<int, int> solutions = GetASolution();
            return solutions;
        }

        public IDictionary<int, int> SolveP(int N)
        {
            IDictionary<int, int> solutions =new Dictionary<int, int>();
            Equation eq = BinarySystem[0];
            if (eq.CountVariables==0)
            {
                if (eq.Value == 0)
                {
                    for (int i=1;i<=N;i++)
                        solutions[i] = 1;
                }
            }
            else
            {
                IList<int> l = eq.Variables.ToList();
                int value = eq.Value;
                for(int i= 1;i<l.Count;i++)
                {
                    solutions[i] = 1;
                    value = (value + 1) % 2;
                }
                solutions[l.Count] = value;
                for (int i = 1; i <= N; i++)
                {
                    if (!solutions.ContainsKey(i))
                        solutions[i] = 1;
                }
            }
            return solutions;

        }
            public bool IsValide()
        {

            for (int l = 0; l < BinarySystem.Count; l++)
            {
                Equation eq = BinarySystem[l];
                if ((eq.CountVariables == 0) && (eq.Value == 1))
                    return false;
            }
            return true;
        }

        private IDictionary<int, int> GetASolution()
        {
            IDictionary<int, int> dic = new Dictionary<int, int>();
            if (!IsValide()) return dic;
            int a;
            int k = 1;

            foreach (Equation v in BinarySystem.Reverse())
            {
                if (v.CountVariables == 0) continue;

                int w = v.Value;
                int first = v.First;
                foreach (int i in v.Variables)
                {
                    if (i == first) continue;

                    if (!dic.ContainsKey(i))
                    {
                        k++;
                        a = 1;
                        dic.Add(i, a);
                        w = a + w;
                    }
                    else
                    {
                        a = dic[i];
                        w = a + w;
                    }
                }
                w = w % 2;
                dic.Add(first, w);
            }
            if (ShowLog)
                DumpSolution(dic);
            return dic;
        }
        public void CheckSolution(IDictionary<int, int> dic)
        {
            if (dic.Count == 0) return;
            foreach (Equation eq in BinarySystem)
            {
                int v = eq.Value;
                int w = 0;
                foreach (int a in eq.Variables)
                {

                    w = dic[a] + w;
                }
                w = w % 2;
                if (v != w)
                {
                    Console.WriteLine(" Trouble with solution");
                    Console.ReadKey();
                }
            }
            Console.WriteLine("solution is Ok");
        }
        public void CountOnes(IDictionary<int, int> dic)
        {
            int n = 0;
            foreach (int i in dic.Values)
            {
                if (i == 1) n++;
            }

            Console.WriteLine(" Number ones = " + n + "  over  " + dic.Count);
        }
        private void DumpSolution(IDictionary<int, int> dic)
        {
            string FileName = Path.Combine(Share, "SOLUTION.txt");

            using (StreamWriter sw = new StreamWriter(FileName))
            {
                string line;
                foreach (int i in dic.Keys)
                {
                    line = i + " = " + dic[i];
                    sw.WriteLine(line);
                }
            }

        }
    }
}

