﻿

using LinearSolver;
using QuadraticSystem;
using SAT2MQ;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Threading;

namespace Fact2SAT
{

    public class SAT
    {
        private IList<Clause> SATClauses = new List<Clause>();
        private IDictionary<int, bool> SATSolutions = new Dictionary<int, bool>();
        private IList<int> Variables = new List<int>();
        private IList<int> NewVariables = new List<int>();
        private LinearSystem BinarySystem = new LinearSystem();
        private Random _random = new Random();
        private int CurrentVariable;
        private Matrix Values = new Matrix();
        private IList<QuadraticEquation> QuadraticSystem = new List<QuadraticEquation>();
        private IDictionary<int, List<int>> VariablesInClauses = new Dictionary<int, List<int>>();
        private int NbrClauses;
        private int NbrVariables;
        public bool Dump = true;
        private readonly string Share = "C:\\Devs\\Output\\SAT";

        public void LoadFile()
        {
            SATClauses.Clear();
            string FileName = Path.Combine(Share, "SAT.txt");

            using (StreamReader sr = new StreamReader(FileName))
            {
                string line;
                while ((line = sr.ReadLine()) != null)
                {
                    string[] data = line.Split(' ');
                    int N = data.Length;
                    Clause cl = new Clause();
                    for (int j = 0; j < N; j++)
                    {
                        int v = int.Parse(data[j]);
                        int k = Math.Abs(v);
                        cl.Add(v);
                        if (!Variables.Contains(k)) Variables.Add(k);
                    }
                    SATClauses.Add(cl);
                }
            }
            // Rename All Variables
            RenameVariables();
        }

        public void LoadFileMQ()
        {
            QuadraticSystem.Clear();
            string FileName = Path.Combine(Share, "MQ.txt");
            using (StreamReader sr = new StreamReader(FileName))
            {
                string line;
                while ((line = sr.ReadLine()) != null)
                {
                    QuadraticEquation qe = new QuadraticEquation();
                    qe.AddTerms(line);
                    QuadraticSystem.Add(qe);
                }
            }
        }
        public void generate3SAT(int clausesNb, int variablesNb)
        {

            Directory.CreateDirectory(Share);

            SATClauses.Clear();
            NbrClauses = clausesNb;
            NbrVariables = variablesNb;
            int N = 3;
            for (int i = 0; i < clausesNb; i++)
            {
                Clause cl = new Clause();
                for (int j = 0; j < N; j++)
                {
                    int k = GetVariable();
                    int v = GetSign() * k;
                    cl.Add(v);
                    if (!Variables.Contains(k)) Variables.Add(k);
                }
                SATClauses.Add(cl);
            }

            // Rename All Variables
            RenameVariables();
            SelectVariableFromSAT();
            if (Dump)
                DumpSat(Path.Combine(Share, "SAT.txt"), SATClauses);
        }

       
        private void Solve2SAT()
        {

        }

        private int GetSign()
        {
            int i = _random.Next(0, 2);
            return i == 0 ? -1 : 1;
        }
        private void SelectVariableFromSAT()
        {
            VariablesInClauses.Clear();
            Variables.Clear();
            int i = 0;
            foreach (Clause cl in SATClauses)
            {
                foreach (int v in cl.Variables)
                {
                    int k = Math.Abs(v);
                    if (!Variables.Contains(k)) Variables.Add(k);

                    if (VariablesInClauses.ContainsKey(v))
                        VariablesInClauses[v].Add(i);
                    else
                        VariablesInClauses.Add(v, new List<int>() { i });

                    i++;
                }
            }

        }
        private int GetVariable()
        {
            return _random.Next(2, NbrClauses);
        }
        private void RenameVariables()
        {
            CurrentVariable = 0;
            List<int> w = Variables.ToList<int>();
            w.Sort();
            for (int i = 0; i < w.Count; i++)
            {
                int v = w[i];
                CurrentVariable++;
                foreach (Clause cl in SATClauses)
                    cl.Rename(v, CurrentVariable);

            }
        }
        public void SAT2MQ()
        {
            int n = GetMaxVariables();
            int m = SATClauses.Count;
            int N = m * (m + 1);
            NewVariables = new List<int>();
            BuildFirstMQ();

            if (CurrentVariable < N)
            {
                AddNewVariablesToMQ(N);
            }
        }
        // add 3* SATClauses.Count new variables if we are in 3SAT
        private void BuildFirstMQ()
        {
            int n = GetMaxVariables();
            int m = SATClauses.Count;
            CurrentVariable = n;
            QuadraticSystem.Clear();
            foreach (Clause cl in SATClauses)
            {
                QuadraticEquation qp = new QuadraticEquation();
                qp.Value = 1;
                for (int i = 0; i < cl.Variables.Count; i++)
                {
                    CurrentVariable++;
                    qp.AddTerms(cl.Variables[i], CurrentVariable);
                    NewVariables.Add(CurrentVariable);

                }
                QuadraticSystem.Add(qp);
            }
        }

        private void AddNewVariablesToMQ(int N)
        {
            for (int i = 0; i < SATClauses.Count; i++)
            {
                Clause cl1 = SATClauses[i];
                foreach (int v in cl1.Variables)
                {
                    for (int j = i + 1; j < SATClauses.Count; j++)
                    {
                        Clause cl2 = SATClauses[i];
                        foreach (int w in cl2.Variables)
                        {
                            if (v == w) continue;
                            CurrentVariable++;
                            QuadraticSystem[i].AddTerms(v, CurrentVariable);
                            QuadraticSystem[j].AddTerms(w, CurrentVariable);
                            if (CurrentVariable > N) return;

                        }
                    }
                }
            }
        }
        private int GetMaxVariables()
        {
            return CurrentVariable;
        }
        public void SolveSystems()
        {
            Stopwatch stopWatch = new Stopwatch();
            stopWatch.Start();

            int N = QuadraticSystem.Count;
            int NN = CountVariables();
            Values.Init(NN, N);

            // 1 =>set all yi*yj coefs to zeros

            Set_Yi_Yj_zeros();

            // 2=> Build and Solve the last system 
            IDictionary<int, int> Y_N_NN = Calc_YN_YNN(N, NN);

            // 3=>Solve Target System;
            IDictionary<int, int> Y_1_N = Calc_Y1_YN(Y_N_NN, N);

            stopWatch.Stop();
            // Get the elapsed time as a TimeSpan value.
            TimeSpan ts = stopWatch.Elapsed;
            string elapsedTime = String.Format("{0:00}:{1:00}:{2:00}.{3:00}",
            ts.Hours, ts.Minutes, ts.Seconds,
            ts.Milliseconds / 10);
            Console.WriteLine("RunTime " + elapsedTime);

            if (Y_1_N.Count == 0)
                Console.WriteLine("Invalid SAT!!!!");
            else
            {
                // calculate output
                IDictionary<int, int> solutions = CalcSolutions(Y_1_N, Y_N_NN);
                Console.WriteLine("Valid SAT!!!!");

            }

            Console.ReadKey();
        }
        private void Set_Yi_Yj_zeros()
        {
            int N = QuadraticSystem.Count;
            int NN = CountVariables();
            for (int i = 2; i <= N; i++)
            {
                BinarySystem.Clear();

                for (int j = 1; j < i; j++)
                {
                    IList<int> hs = Values.GetColumn(j);
                    HashSet<Equation> bs = GetLinearSystem(i, hs);
                    BinarySystem.AddBinarySystem(bs);
                }

                if (i % 5 == 0)
                {
                    Console.WriteLine(i + " binary systems solved. Time = " + DateTime.Now);
                    Console.WriteLine(BinarySystem.Count + " binary systems size. Time = " + DateTime.Now);
                }

                BinarySystem.UpdateSystem(i, NN);
                IDictionary<int, int> solutions = BinarySystem.SolveP();
                // Add default values
                for (int j = 1; j <= NN; j++)
                {
                    if (solutions.ContainsKey(j)) continue;
                    if (i + j == NN + 1)
                        solutions[j] = 1;
                    else if (i + j > NN + 1)
                        solutions[j] = 0;
                    else
                        solutions[j] = 1;
                }

                Values.UpdateCol(i, solutions);
            }

        }

        private HashSet<Equation> GetLinearSystem(int indexRow, IList<int> hs)
        {
            HashSet<Equation> sys = new HashSet<Equation>();
            foreach (QuadraticEquation eq in QuadraticSystem)
            {
                Equation leq = new Equation();
                leq.Value = 0;
                foreach (Term trn in eq.Terms)
                {
                    int a = trn.First;
                    int b = trn.Second;
                    int r = hs[a - 1];
                    int s = hs[b - 1];
                    if (r == 1) leq.Add(b);
                    if (s == 1) leq.Add(a);
                }
                sys.Add(leq);
            }
            return sys;
        }
        private IDictionary<int, int> Calc_YN_YNN(int N, int NN)
        {
            GetLastSystem();
            IDictionary<int, int> Y_N_NN = BinarySystem.SolveP();
            CompleteSolutions(Y_N_NN, NN);
            return Y_N_NN;
        }

        private IDictionary<int, int> Calc_Y1_YN(IDictionary<int, int> Y_N_NN, int N)
        {
            IDictionary<int, int> Y_1_N;
            BuildTargetSystem(Y_N_NN);
            if (BinarySystem.Count == 1)
                Y_1_N = BinarySystem.SolveP(N);
            else
                Y_1_N = BinarySystem.SolveP();
            if (Y_1_N.Count == 0) return Y_1_N;
            CompleteSolutions(Y_1_N, N);
            return Y_1_N;
        }
        private void CompleteSolutions(IDictionary<int, int> dic, int Nb)
        {
            int N = QuadraticSystem.Count;
            for (int i = 1; i <= Nb; i++)
            {
                if (!dic.ContainsKey(i))
                    dic[i] = 1;
            }
        }

        private IDictionary<int, int> CalcSolutions(IDictionary<int, int> Y_1_N, IDictionary<int, int> Y_N_NN)
        {

            IDictionary<int, int> solutions = new Dictionary<int, int>();
            int N = QuadraticSystem.Count;
            int NN = CountVariables();
            for (int i = 1; i <= NN; i++)
            {
                int value = 0;
                for (int j = 1; j <= NN; j++)
                {
                    int a = (j <= N) ? Y_1_N[j] : Y_N_NN[j];
                    int b = Values.Getvalue(i, j);
                    value = (value + a * b) % 2;
                }
                solutions.Add(i, value);
            }
            return solutions;
        }
        private void GetLastSystem()
        {
            BinarySystem.Clear();
            foreach (QuadraticEquation qp in QuadraticSystem)
                GetLastLinearSystem(qp);

        }
        private void BuildTargetSystem(IDictionary<int, int> Y_N_NN)
        {
            BinarySystem.Clear();
            foreach (QuadraticEquation qp in QuadraticSystem)
                GetTargetLinearSystem(qp, Y_N_NN);
               
           

        }
        private void GetTargetLinearSystem(QuadraticEquation qp, IDictionary<int, int> Y_N_NN)
        {

            Equation le = new Equation();
            int N = QuadraticSystem.Count;
            int NN = CountVariables();
            le.Value = 1;
            le.Value = (le.Value + CalcQ(qp, Y_N_NN, N, NN)) % 2;
            foreach (Term trm in qp.Terms)
            {
                int a = trm.First;
                int b = trm.Second;
                for (int i = 1; i <= N; i++)
                {
                    int r = Values.Getvalue(a, i);
                    int s = Values.Getvalue(b, i);
                    if (r*s == 1)
                        le.Add(i);
                }
            }
            BinarySystem.Add(le);
        }

        private int CalcQ(QuadraticEquation qp, IDictionary<int, int> Y_N_NN, int N, int NN)
        {
            int value = 0;

            foreach (Term trm in qp.Terms)
            {
                int value1 = 0;
                int value2 = 0;
                int a = trm.First;
                int b = trm.Second;
                for (int i = N + 1; i <= NN; i++)
                {
                    int r = Values.Getvalue(a, i);
                    int s = Values.Getvalue(b, i);
                    value1 = (value1 + r * Y_N_NN[i]) % 2;
                    value2 = (value2 + s * Y_N_NN[i]) % 2;
                }

                value = (value + value1 * value2) % 2;
            }
            return value;
        }
        private void GetLastLinearSystem(QuadraticEquation qp)
        {
            HashSet<Equation> sys = new HashSet<Equation>();
            int N = QuadraticSystem.Count;
            int NN = CountVariables();
            for (int i = 1; i <= N; i++)
            {
                Equation leq = new Equation();
                leq.Value = 0;

                foreach (Term trn in qp.Terms)
                {
                    int a = trn.First;
                    int b = trn.Second;
                    int r = Values.Getvalue(a, i);
                    int s = Values.Getvalue(b, i);
                    if (a == b)
                    {
                        if (s == 1)
                            leq.Value = (leq.Value + 1) % 2;
                        continue;
                    }



                    if (r == 1)
                    {
                        for (int j = N + 1; j <= NN; j++)
                        {
                            int value = Values.Getvalue(b, j);
                            if (value == 1)
                                leq.Add(j);
                        }
                    }

                    if (s == 1)
                    {
                        for (int j = N + 1; j <= NN; j++)
                        {
                            int value = Values.Getvalue(a, j);
                            if (value == 1)
                                leq.Add(j);
                        }
                    }

                }
                if (leq.CountVariables > 0)
                    sys.Add(leq);
            }

            foreach (Equation eq in sys)
                BinarySystem.Add(eq);


        }

        private int CountVariables()
        {
            return CurrentVariable; //+ NewVariables.Count;
        }

        private void DumpSat(string fileName, IList<Clause> clauses)
        {
            int count = 0;

            using (StreamWriter sw = new StreamWriter(fileName))
            {

                foreach (Clause cl in clauses)
                {
                    string line = cl.GetLine();
                    sw.WriteLine(line);
                    count++;
                }

            }
            Console.WriteLine(count + " =  lines are saved to the file.");
        }
    }
}
